from typing import List, Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time


class AverageReqResourceOptimizer(ResourceOptimizer):

    def __init__(self, k: float = 2):
        self.k = k

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: List[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[List[Worker]], Time]):

        if optimize_array:
            for i in range(len(worker_team)):
                if optimize_array[i]:
                    worker_team[i].count = max(1, down_border[i]) + int((up_border[i] - down_border[i]) / self.k)
        else:
            # TODO Remove max()
            for i in range(len(worker_team)):
                worker_team[i].count = max(1, down_border[i]) + int((up_border[i] - down_border[i]) / self.k)



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, Callable

from sampo.scheduler.base import Scheduler
from sampo.scheduler.multi_agency.block_graph import BlockGraph
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.utils.obstruction import Obstruction
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.time import Time


class Agent:

    def __init__(self, name: str, scheduler: Scheduler, contractors: list[Contractor]):
        self.name = name
        self._timeline = None
        self._scheduler = scheduler
        self._contractors = contractors
        self._last_task_executed = Time(0)
        self._downtime = Time(0)

    def offer(self, wg: WorkGraph, parent_time: Time) -> tuple[Time, Time, Schedule, Timeline]:
        """
        Computes the offer from agent to manager. Handles all works from given wg.

        :param wg: the given block of tasks
        :param parent_time: max end time of parent blocks
        :return: offered start time, end time, resulting schedule and timeline before offering

        To apply returned offer, use `Agent#confirm`.
        """
        schedule, start_time, timeline, _ = \
            self._scheduler.schedule_with_cache(wg, self._contractors,
                                                assigned_parent_time=parent_time, timeline=deepcopy(self._timeline))
        return start_time, schedule.execution_time, schedule, timeline

    def confirm(self, timeline: Timeline, start: Time, end: Time):
        """
        Applies the given offer.

        :param timeline: timeline returned from corresponding `Agent#offer`
        :param start: global start time of confirmed block
        :param end: global end time of confirmed block
        """
        self._timeline = timeline
        self.update_stat(start)
        # update last task statistic
        self._last_task_executed = end

    def update_stat(self, start: Time):
        # count last iteration downtime
        # if given start time is lower than last executed task
        #   then this downtime are already in self_downtime
        self._downtime += max(Time(0), start - self._last_task_executed)

    def __str__(self):
        return f'Agent(name={self.name}, scheduler={self._scheduler}, downtime={self._downtime})'

    def __repr__(self):
        return str(self)

    @property
    def downtime(self) -> Time:
        return self._downtime

    @property
    def contractors(self):
        return self._contractors

    @property
    def scheduler(self):
        return self._scheduler


@dataclass
class ScheduledBlock:
    """
    An object represents scheduled graph block(group of works).

    Contains all data used in scheduling, the agent and resulting information.
    """
    wg: WorkGraph
    schedule: Schedule
    agent: Agent
    start_time: Time
    end_time: Time

    @property
    def id(self):
        return self.wg.start.id

    @property
    def duration(self):
        return self.end_time - self.start_time

    def __str__(self):
        return f'ScheduledBlock(start_time={self.start_time}, end_time={self.end_time}, agent={self.agent})'

    def __repr__(self):
        return str(self)


class Manager:
    def __init__(self, agents: list[Agent]):
        if len(agents) == 0:
            raise Exception("Manager can't work with empty list of agents")
        self._agents = agents

    # TODO Upgrade to supply the best parallelism
    def manage_blocks(self, bg: BlockGraph, logger: Callable[[str], None] = None) -> Dict[str, ScheduledBlock]:
        """
        Runs multi-agent system based on auction on given BlockGraph.
        
        :param bg: 
        :param logger:
        :return: an index of resulting `ScheduledBlock`s built by ids of corresponding `WorkGraph`s
        """
        id2sblock = {}
        for i, block in enumerate(bg.toposort()):
            max_parent_time = max((id2sblock[parent.id].end_time for parent in block.blocks_from), default=Time(0)) + 1
            start_time, end_time, agent_schedule, agent \
                = self.run_auction_with_obstructions(block.wg, max_parent_time, block.obstruction)

            assert start_time >= max_parent_time, f'Scheduler {agent._scheduler} does not handle parent_time!'

            if logger:
                logger(f'{agent._scheduler}')
            sblock = ScheduledBlock(wg=block.wg, agent=agent, schedule=agent_schedule,
                                    start_time=start_time,
                                    end_time=end_time)
            id2sblock[sblock.id] = sblock

        return id2sblock

    def run_auction_with_obstructions(self, wg: WorkGraph, parent_time: Time = Time(0),
                                      obstruction: Obstruction | None = None):
        if obstruction:
            obstruction.generate(wg)
        return self.run_auction(wg, parent_time)

    def run_auction(self, wg: WorkGraph, parent_time: Time = Time(0)) -> (Time, Time, Schedule, Agent):
        """
        Runs the auction on the given `WorkGraph`.

        :param wg: target `WorkGraph`
        :param parent_time: max parent time of given block
        :return: best start time, end time and the agent that is able to support this working time
        """
        best_start_time = 0
        best_end_time = Time.inf()
        best_schedule = None
        best_timeline = None
        best_agent = None

        offers = [(agent, agent.offer(wg, parent_time)) for agent in self._agents]

        for offered_agent, (offered_start_time, offered_end_time, offered_schedule, offered_timeline) in offers:
            if offered_end_time < best_end_time:
                best_start_time = offered_start_time
                best_end_time = offered_end_time
                best_schedule = offered_schedule
                best_timeline = offered_timeline
                best_agent = offered_agent
        best_agent.confirm(best_timeline, best_start_time, best_end_time)
        for agent in self._agents:
            if agent.name != best_agent.name:
                agent.update_stat(best_start_time)

        return best_start_time, best_end_time, best_schedule, best_agent



from enum import Enum
from random import Random
from typing import Callable
from uuid import uuid4

from sampo.generator import SimpleSynthetic
from sampo.generator.pipeline.types import SyntheticGraphType
from sampo.scheduler.multi_agency.block_graph import BlockGraph, BlockNode
from sampo.scheduler.utils.obstruction import Obstruction
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.works import WorkUnit


class SyntheticBlockGraphType(Enum):
    Sequential = 0,
    Parallel = 1,
    Random = 2,
    Queues = 3


EMPTY_GRAPH_VERTEX_COUNT = 2


def generate_empty_graph() -> WorkGraph:
    start = GraphNode(WorkUnit(str(uuid4()), ""), [])
    end = GraphNode(WorkUnit(str(uuid4()), ""), [start])
    return WorkGraph(start, end)


def generate_blocks(graph_type: SyntheticBlockGraphType, n_blocks: int, type_prop: list[int],
                    count_supplier: Callable[[int], tuple[int, int]],
                    edge_prob: float, rand: Random | None = Random(),
                    obstruction_getter: Callable[[int], Obstruction | None] = lambda _: None,
                    logger: Callable[[str], None] = print) -> BlockGraph:
    """
    Generate synthetic block graph according to given parameters.

    :param graph_type: type of BlockGraph
    :param n_blocks: the count of blocks
    :param type_prop: proportions of the `WorkGraph` types: General, Parallel, Sequential
    :param count_supplier: function that computes the borders of block size from it's index
    :param edge_prob: edge existence probability
    :param rand: a random reference
    :param obstruction_getter:
    :return: generated block graph
    """
    ss = SimpleSynthetic(rand)

    modes = rand.sample(list(SyntheticGraphType), counts=[p * n_blocks for p in type_prop], k=n_blocks)
    nodes = [ss.work_graph(mode, *count_supplier(i)) for i, mode in enumerate(modes)]
    nodes += [generate_empty_graph(), generate_empty_graph()]
    bg = BlockGraph.pure(nodes, obstruction_getter)

    global_start, global_end = bg.nodes[-2:]

    match graph_type:
        case SyntheticBlockGraphType.Sequential:
            for idx, start in enumerate(bg.nodes[:-2]):
                if start.wg.vertex_count > EMPTY_GRAPH_VERTEX_COUNT \
                        and bg.nodes[idx + 1].wg.vertex_count > EMPTY_GRAPH_VERTEX_COUNT:
                    bg.add_edge(start, bg.nodes[idx + 1])
        case SyntheticBlockGraphType.Parallel:
            pass
        case SyntheticBlockGraphType.Random:
            rev_edge_prob = int(1 / edge_prob)
            for idx, start in enumerate(bg.nodes):
                for end in bg.nodes[idx:]:
                    if start.wg.vertex_count > EMPTY_GRAPH_VERTEX_COUNT \
                            and end.wg.vertex_count > EMPTY_GRAPH_VERTEX_COUNT:
                        if start == end or rand.randint(0, rev_edge_prob) != 0:
                            continue
                        bg.add_edge(start, end)

    for node in bg.nodes:
        if node.wg.vertex_count > EMPTY_GRAPH_VERTEX_COUNT:
            bg.add_edge(global_start, node)
            bg.add_edge(node, global_end)

    logger(f'{graph_type.name} ' + ' '.join([str(mode.name) for mode in modes]))
    return bg


def generate_block_graph(graph_type: SyntheticBlockGraphType, n_blocks: int, type_prop: list[int],
                         count_supplier: Callable[[int], tuple[int, int]],
                         edge_prob: float, rand: Random | None = Random(),
                         obstruction_getter: Callable[[int], Obstruction | None] = lambda _: None,
                         queues_num: int | None = None,
                         queues_blocks: list[int] | None = None,
                         queues_edges: list[int] | None = None,
                         logger: Callable[[str], None] = print) -> BlockGraph:
    if graph_type == SyntheticBlockGraphType.Queues:
        return generate_queues(type_prop, count_supplier, rand, obstruction_getter, queues_num, queues_blocks,
                               queues_edges, logger)
    else:
        return generate_blocks(graph_type, n_blocks, type_prop, count_supplier, edge_prob, rand, obstruction_getter,
                               logger)


def generate_queues(type_prop: list[int],
                    count_supplier: Callable[[int], tuple[int, int]],
                    rand: Random | None = Random(),
                    obstruction_getter: Callable[[int], Obstruction | None] = lambda _: None,
                    queues_num: int | None = None,
                    queues_blocks: list[int] | None = None,
                    queues_edges: list[int] | None = None,
                    logger: Callable[[str], None] = print) -> BlockGraph:
    """
    Generate synthetic block queues graph according to given parameters.

    :param type_prop: proportions of the `WorkGraph` types: General, Parallel, Sequential
    :param count_supplier: function that computes the borders of block size from it's index
    :param rand: a random reference
    :param obstruction_getter:
    :param queues_num:
    :param queues_blocks:
    :param queues_edges:
    :return: generated block graph
    """
    ss = SimpleSynthetic(rand)
    nodes_all: list[BlockNode] = []
    nodes_prev: list[BlockNode] = []

    for queue in range(queues_num):
        # generate vertices
        n_blocks = queues_blocks[queue]
        modes = rand.sample(list(SyntheticGraphType), counts=[p * n_blocks for p in type_prop], k=n_blocks)
        nodes = [BlockNode(ss.work_graph(mode, *count_supplier(i)), obstruction_getter(i))
                 for i, mode in enumerate(modes)]
        nodes_all.extend(nodes)
        if not nodes_all:
            nodes_prev = nodes
            # logger(f'Generated queue 0: blocks={n_blocks}')
            continue

        # generate edges
        generated_edges = 0
        for i, node in enumerate(nodes[:-2]):
            if i >= len(nodes_prev):
                break
            BlockGraph.add_edge(node, nodes_prev[i])
            generated_edges += 1

        for i, node in enumerate(nodes):
            if i >= len(nodes_prev):
                break
            # we are going in reverse to fill edges that are not covered by previous cycle
            BlockGraph.add_edge(node, nodes_prev[-i])
            generated_edges += 1

        nodes_prev = nodes

        # logger(f'Generated queue {queue}: blocks={n_blocks}, edges={generated_edges}')
    logger('Queues')

    return BlockGraph(nodes_all)





import queue

from sampo.schemas.graph import GraphNode


def count_node_ancestors(finish: GraphNode, root: GraphNode) -> int:
    q = queue.Queue()
    count = 0
    used = set()
    used.add(root)
    q.put(finish)
    while not q.empty():
        node = q.get()
        for parent in node.parents:
            if parent in used:
                continue
            used.add(parent)
            q.put(parent)
            count += 1

    return count



from typing import List, Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.utilities.base_opt import dichotomy_int


class FullScanResourceOptimizer(ResourceOptimizer):

    _coordinate_descent_optimizer = CoordinateDescentResourceOptimizer(dichotomy_int)

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: List[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[List[Worker]], Time]):

        # TODO Handle optimize_array
        def fitness(worker_count: np.ndarray):
            for worker_ind in range(len(worker_team)):
                worker_team[worker_ind].count = worker_count[worker_ind]
            return get_finish_time(worker_team)

        right_fit = fitness(up_border)
        left_fit = fitness(down_border)

        m = up_border

        # Trying to +1 to all workers
        while (m > down_border).any():
            m = (up_border + down_border) // 2
            next_fit = fitness(m)
            if next_fit < right_fit:
                down_border = m
                left_fit = next_fit
            elif next_fit > left_fit:
                up_border = m
                right_fit = next_fit
            else:
                break

        # Insert cur into coordinate-descent optimizer as down_border
        self._coordinate_descent_optimizer.optimize_resources(worker_pool,
                                                              worker_team,
                                                              optimize_array,
                                                              m,
                                                              up_border,
                                                              get_finish_time)



from dataclasses import dataclass
from random import Random
from typing import Optional

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.interval import IntervalGaussian
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.types import AgentId


@dataclass
class Resource(AutoJSONSerializable['Equipment'], Identifiable):
    """
    A class summarizing the different resources used in the work: Human resources, equipment, materials, etc.
    """
    pass


# TODO describe "productivity"
@dataclass
class Worker(Resource):
    """
    A class dedicated to human resources
    :param count: the number of people in this resource
    :param contractor_id: Contractor id if resources are added directly to the contractor
    :param productivity: Contractor id if resources are added directly to the contractor
    """

    def __init__(self,
                 id: str,
                 name: str,
                 count: int,
                 contractor_id: Optional[str] = "",
                 productivity: Optional[IntervalGaussian] = IntervalGaussian(1, 0, 1, 1),
                 cost_one_unit: Optional[float] = None):
        super(Worker, self).__init__(id, name)
        self.count = count
        self.contractor_id = contractor_id
        self.productivity = productivity if productivity is not None else IntervalGaussian(1, 0, 1, 1)
        self.cost_one_unit = cost_one_unit if cost_one_unit is not None else self.productivity.mean * 10

    ignored_fields = ['productivity']

    # TODO: describe the function (description, return type)
    def copy(self):
        return Worker(id=self.id,
                      name=self.name,
                      count=self.count,
                      contractor_id=self.contractor_id,
                      productivity=self.productivity)

    def with_count(self, count: int) -> 'Worker':
        self.count = count
        return self

    def get_cost(self) -> float:
        """Returns cost of this worker entry"""
        return self.cost_one_unit * int(self.count)

    # TODO: describe the function (description, return type)
    def get_agent_id(self) -> AgentId:
        return self.contractor_id, self.name

    # TODO: describe the function (description, return type)
    def get_static_productivity(self) -> float:
        return self.productivity.mean * self.count

    # TODO: describe the function (description, return type)
    def get_stochastic_productivity(self, rand: Optional[Random] = None) -> float:
        return self.productivity.rand_float(rand) * self.count

    # TODO: describe the function (description, return type)
    def __repr__(self):
        return f'{self.count} {self.name}'

# TODO: describe the class (description)
@dataclass
class ConstructionObject(Resource):
    pass

# TODO: describe the class (description, parameters)
@dataclass(init=False)
class EmptySpaceConstructionObject(ConstructionObject):
    id: str = "00000000000000000"
    name: str = "empty space construction object"

# TODO: describe the class (description)
@dataclass
class Equipment(Resource):
    pass

# TODO: describe the class (description)
@dataclass
class Material(Resource):
    pass



import math
import random
import time
from typing import Dict, List, Tuple, Callable

import numpy as np
import seaborn as sns
from deap import tools
from deap.tools import initRepeat
from matplotlib import pyplot as plt
from pandas import DataFrame

from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome, convert_chromosome_to_schedule
from sampo.scheduler.genetic.operators import init_toolbox, ChromosomeType, Individual, copy_chromosome, \
    FitnessFunction, TimeFitness, is_chromosome_correct
from sampo.scheduler.native_wrapper import NativeWrapper
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.schedule import ScheduleWorkDict, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.collections_util import reverse_dictionary


def build_schedule(wg: WorkGraph,
                   contractors: List[Contractor],
                   worker_pool: WorkerContractorPool,
                   population_size: int,
                   generation_number: int,
                   selection_size: int,
                   mutate_order: float,
                   mutate_resources: float,
                   init_schedules: Dict[str, tuple[Schedule, list[GraphNode] | None]],
                   rand: random.Random,
                   spec: ScheduleSpec,
                   fitness_constructor: Callable[[Callable[[list[ChromosomeType]], list[int]]],
                                                 FitnessFunction] = TimeFitness,
                   work_estimator: WorkTimeEstimator = None,
                   show_fitness_graph: bool = False,
                   n_cpu: int = 1,
                   assigned_parent_time: Time = Time(0),
                   timeline: Timeline | None = None,
                   time_border: int = None) \
        -> tuple[ScheduleWorkDict, Time, Timeline, list[GraphNode]]:
    """
    Genetic algorithm
    Structure of chromosome:
    [[order of job], [numbers of workers types 1 for each job], [numbers of workers types 2], ... ]
    Different mate and mutation for order and for workers
    Generate order of job by prioritization from HEFTs and from Topological
    Generate resources from min to max
    Overall initial population is valid

    :param fitness_constructor:
    :param show_fitness_graph:
    :param worker_pool:
    :param contractors:
    :param wg:
    :param population_size:
    :param generation_number:
    :param selection_size:
    :param mutate_order:
    :param mutate_resources:
    :param rand:
    :param spec: spec for current scheduling
    :param init_schedules:
    :param timeline:
    :param n_cpu: number or parallel workers to use in computational process
    :param assigned_parent_time: start time of the whole schedule(time shift)
    :param work_estimator:
    :param time_border:
    :return: scheduler
    """

    if show_fitness_graph:
        fitness_history = list()

    global_start = time.time()

    start = time.time()
    # preparing access-optimized data structures
    nodes = [node for node in wg.nodes if not node.is_inseparable_son()]

    index2node: Dict[int, GraphNode] = {index: node for index, node in enumerate(nodes)}
    work_id2index: Dict[str, int] = {node.id: index for index, node in index2node.items()}
    worker_name2index = {worker_name: index for index, worker_name in enumerate(worker_pool)}
    index2contractor = {ind: contractor.id for ind, contractor in enumerate(contractors)}
    index2contractor_obj = {ind: contractor for ind, contractor in enumerate(contractors)}
    contractor2index = reverse_dictionary(index2contractor)
    index2node_list = [(index, node) for index, node in enumerate(nodes)]
    worker_pool_indices = {worker_name2index[worker_name]: {
        contractor2index[contractor_id]: worker for contractor_id, worker in workers_of_type.items()
    } for worker_name, workers_of_type in worker_pool.items()}
    node_indices = list(range(len(nodes)))

    contractors_capacity = np.zeros((len(contractors), len(worker_pool)))
    for w_ind, cont2worker in worker_pool_indices.items():
        for c_ind, worker in cont2worker.items():
            contractors_capacity[c_ind][w_ind] = worker.count

    resources_border = np.zeros((2, len(worker_pool), len(index2node)))
    resources_min_border = np.zeros((len(worker_pool)))
    for work_index, node in index2node.items():
        for req in node.work_unit.worker_reqs:
            worker_index = worker_name2index[req.kind]
            resources_border[0, worker_index, work_index] = req.min_count
            resources_border[1, worker_index, work_index] = req.max_count
            resources_min_border[worker_index] = max(resources_min_border[worker_index], req.min_count)

    contractor_borders = np.zeros((len(contractor2index), len(worker_name2index)), dtype=int)
    for ind, contractor in enumerate(contractors):
        for ind_worker, worker in enumerate(contractor.workers.values()):
            contractor_borders[ind, ind_worker] = worker.count

    # construct inseparable_child -> inseparable_parent mapping
    inseparable_parents = {}
    for node in nodes:
        for child in node.get_inseparable_chain_with_self():
            inseparable_parents[child] = node

    # here we aggregate information about relationships from the whole inseparable chain
    children = {work_id2index[node.id]: [work_id2index[inseparable_parents[child].id]
                                         for inseparable in node.get_inseparable_chain_with_self()
                                         for child in inseparable.children]
                for node in nodes}

    parents = {work_id2index[node.id]: [] for node in nodes}
    for node, node_children in children.items():
        for child in node_children:
            parents[child].append(node)

    print(f'Genetic optimizing took {(time.time() - start) * 1000} ms')

    start = time.time()

    # initial chromosomes construction
    init_chromosomes: Dict[str, ChromosomeType] = \
        {name: convert_schedule_to_chromosome(wg, work_id2index, worker_name2index,
                                              contractor2index, contractor_borders, schedule, order)
         for name, (schedule, order) in init_schedules.items()}

    toolbox = init_toolbox(wg, contractors, worker_pool, index2node,
                           work_id2index, worker_name2index, index2contractor,
                           index2contractor_obj, init_chromosomes, mutate_order,
                           mutate_resources, selection_size, rand, spec, worker_pool_indices,
                           contractor2index, contractor_borders, node_indices, index2node_list, parents,
                           assigned_parent_time, work_estimator)

    for name, chromosome in init_chromosomes.items():
        if not is_chromosome_correct(chromosome, node_indices, parents):
            raise NoSufficientContractorError('HEFTs are deploying wrong chromosomes')

    # save best individuals
    hof = tools.HallOfFame(1, similar=compare_individuals)
    # create population of a given size
    pop = toolbox.population(n=population_size)

    # probability to participate in mutation and crossover for each individual
    cxpb, mutpb = mutate_order, mutate_order
    mutpb_res, cxpb_res = mutate_resources, mutate_resources

    native = NativeWrapper(toolbox, wg, contractors, worker_name2index, worker_pool_indices, work_estimator)

    # def evaluate_chromosomes(chromosomes: list[ChromosomeType]):
    #     return native.evaluate([chromo for chromo in chromosomes if toolbox.validate(chromo)])

    fitness_f = fitness_constructor(native.evaluate)

    print(f'Toolbox initialization & first population took {(time.time() - start) * 1000} ms')
    start = time.time()

    # map to each individual fitness function
    pop = [ind for ind in pop if toolbox.validate(ind[0])]
    fitness = fitness_f.evaluate([ind[0] for ind in pop])

    for ind, fit in zip(pop, fitness):
        ind.fitness.values = [fit]
        ind.fitness.invalid_steps = 1 if fit == Time.inf() else 0

    hof.update(pop)
    best_fitness = hof[0].fitness.values[0]

    if show_fitness_graph:
        fitness_history.append(sum(fitness) / len(fitness))

    g = 0
    # the best fitness, track to increase performance by stopping evaluation when not decreasing
    prev_best_fitness = Time.inf()

    print(f'First population evaluation took {(time.time() - start) * 1000} ms')
    start = time.time()

    plateau_steps = 0
    max_plateau_steps = 8

    while g < generation_number and plateau_steps < max_plateau_steps \
            and (time_border is None or time.time() - global_start < time_border):
        print(f"-- Generation {g}, population={len(pop)}, best time={best_fitness} --")
        if best_fitness == prev_best_fitness:
            plateau_steps += 1
        else:
            plateau_steps = 0
        prev_best_fitness = best_fitness

        # select individuals of next generation
        offspring = toolbox.select(pop, int(math.sqrt(len(pop))))
        # clone selected individuals
        # offspring = [toolbox.clone(ind) for ind in offspring]

        # operations for ORDER
        # crossover
        # take 2 individuals as input 1 modified individuals
        # take after 1: (1,3,5) and (2,4,6) and get pairs 1,2; 3,4; 5,6

        cur_generation = []

        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            if rand.random() < cxpb:
                ind1, ind2 = toolbox.mate(child1[0], child2[0])
                # add to population
                cur_generation.append(wrap(ind1))
                cur_generation.append(wrap(ind2))

        # mutation
        # take 1 individuals as input and return 1 individuals as output
        for mutant in offspring:
            if rand.random() < mutpb:
                ind_order = toolbox.mutate(mutant[0][0])
                ind = copy_chromosome(mutant[0])
                ind = (ind_order[0], ind[1], ind[2])
                # add to population
                cur_generation.append(wrap(ind))

        # operations for RESOURCES
        # mutation
        # select types for mutation
        # numbers of changing types
        number_of_type_for_changing = rand.randint(1, len(worker_name2index) - 1)
        # workers type for changing(+1 means contractor 'resource')
        workers = rand.sample(range(len(worker_name2index) + 1), number_of_type_for_changing)

        # resources mutation
        for worker in workers:
            low = resources_border[0, worker] if worker != len(worker_name2index) else 0
            up = resources_border[1, worker] if worker != len(worker_name2index) else 0
            for mutant in offspring:
                if rand.random() < mutpb_res:
                    ind = toolbox.mutate_resources(mutant[0], low=low, up=up, type_of_worker=worker)
                    # add to population
                    cur_generation.append(wrap(ind))

        # resource borders mutation
        for worker in workers:
            if worker == len(worker_name2index):
                continue
            for mutant in offspring:
                if rand.random() < mutpb_res:
                    ind = toolbox.mutate_resource_borders(mutant[0],
                                                          contractors_capacity=contractors_capacity,
                                                          resources_min_border=resources_min_border,
                                                          type_of_worker=worker)
                    # add to population
                    cur_generation.append(wrap(ind))

        # for the crossover, we use those types that did not participate in the mutation(+1 means contractor 'resource')
        # workers_for_mate = list(set(list(range(len(worker_name2index) + 1))) - set(workers))
        # crossover
        # take 2 individuals as input 1 modified individuals

        workers = rand.sample(range(len(worker_name2index) + 1), number_of_type_for_changing)

        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            for ind_worker in workers:
                # mate resources
                if rand.random() < cxpb_res:
                    ind1, ind2 = toolbox.mate_resources(child1[0], child2[0], ind_worker)
                    # add to population
                    cur_generation.append(wrap(ind1))
                    cur_generation.append(wrap(ind2))

                # mate resource borders
                if rand.random() < cxpb_res:
                    if ind_worker == len(worker_name2index):
                        continue
                    ind1, ind2 = toolbox.mate_resource_borders(child1[0], child2[0], ind_worker)

                    # add to population
                    cur_generation.append(wrap(ind1))
                    cur_generation.append(wrap(ind2))

        # add mutant part of generation to offspring
        offspring.extend(cur_generation)
        cur_generation.clear()
        # Gather all the fitness in one list and print the stats
        invalid_ind = [ind for ind in offspring if toolbox.validate(ind[0])]
        # for each individual - evaluation
        # print(pool.map(lambda x: x + 2, range(10)))

        invalid_fit = fitness_f.evaluate([ind[0] for ind in invalid_ind if toolbox.validate(ind[0])])
        for fit, ind in zip(invalid_fit, invalid_ind):
            ind.fitness.values = [fit]
            if fit == Time.inf() and ind.fitness.invalid_steps == 0:
                ind.fitness.invalid_steps = 1

        if show_fitness_graph:
            _ftn = [f for f in fitness if not math.isinf(f)]
            if len(_ftn) > 0:
                fitness_history.append(sum(_ftn) / len(_ftn))

        # pop_size = len(pop)
        # pop = [ind for ind in pop if valid(ind)]
        # print(f'----| Filtered out {pop_size - len(pop)} invalid individuals')

        # renewing population
        pop[:] = offspring
        hof.update(pop)

        best_fitness = hof[0].fitness.values[0]

        # best = hof[0]
        # fits = [ind.fitness.values[0] for ind in pop]
        # evaluation = chromosome_evaluation(best, index2node, resources_border, work_id2index, worker_name2index,
        #                                   parent2inseparable_son, agents)
        # print("fits: ", fits)
        # print(evaluation)
        g += 1

    native.close()

    chromosome = hof[0][0]

    # assert that we have valid chromosome
    assert hof[0].fitness.values[0] != Time.inf()

    print(f'Final time: {hof[0].fitness.values[0]}')

    scheduled_works, schedule_start_time, timeline, order_nodes \
        = convert_chromosome_to_schedule(chromosome, worker_pool, index2node,
                                         index2contractor_obj,
                                         worker_pool_indices,
                                         spec, timeline,
                                         assigned_parent_time,
                                         work_estimator)

    print(f'Generations processing took {(time.time() - start) * 1000} ms')

    if show_fitness_graph:
        sns.lineplot(
            data=DataFrame.from_records([(g * 4, v) for g, v in enumerate(fitness_history)],
                                        columns=["Поколение", "Функция качества"]),
            x="Поколение",
            y="Функция качества",
            palette='r')
        plt.show()

    return {node.id: work for node, work in scheduled_works.items()}, schedule_start_time, timeline, order_nodes


def compare_individuals(a: Tuple[ChromosomeType], b: Tuple[ChromosomeType]):
    return (a[0][0] == b[0][0]).all() and (a[0][1] == b[0][1]).all()


def wrap(chromosome: ChromosomeType) -> Individual:
    """
    Created an individual from chromosome

    :param chromosome:
    :return:
    """

    def ind_getter():
        return chromosome

    ind = initRepeat(Individual, ind_getter, n=1)
    ind.fitness.invalid_steps = 0
    return ind



from dataclasses import dataclass
from enum import Enum
from typing import Tuple, Optional

from sampo.schemas.time import Time

ContractorName = str
WorkerName = str
AgentId = Tuple[ContractorName, WorkerName]


# TODO check for the possibility of translation to IntEnum with the removal of priority
# TODO: describe the class (description)
class EventType(Enum):
    Initial = -1
    Start = 0
    End = 1

    @property
    def priority(self) -> int:
        """
        Returns the processing order for scheduling events

        :return value: The desired priority
        """
        # noinspection PyTypeChecker
        value = int(self.value)
        return value


# TODO: describe the class (description, parameters)
@dataclass
class ScheduleEvent:
    seq_id: int
    event_type: EventType
    time: Time
    swork: Optional['ScheduledWork']
    available_workers_count: int



import os
import sys


def get_absolute_resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath("")

    return os.path.join(base_path, relative_path)
